@@ -2,12 +2,17 @@ import asyncio
 import inspect
 import json
 import logging
+import time
 import traceback
 import uuid
 from asyncio import Queue, CancelledError
 from os import environ
 
 import aiohttp
+import requests
+
+from metrics.constants import OUTGOING_REQUEST_LATENCY_SEC
+from rasa.core.utils import AvailableEndpoints
 from sanic import Sanic, Blueprint, response
 from sanic.request import Request
 from typing import Text, List, Dict, Any, Optional, Callable, Iterable, Awaitable
@@ -16,6 +21,7 @@ import rasa.utils.endpoints
 from rasa.cli import utils as cli_utils
 from rasa.constants import DOCS_BASE_URL
 from rasa.core import utils
+from sanic.response import HTTPResponse
 
 try:
     from urlparse import urljoin  # pytype: disable=import-error
@@ -23,9 +29,9 @@ except ImportError:
     from urllib.parse import urljoin
 
 logger = logging.getLogger(__name__)
+OMNI_API_URL = environ.get("OMNI_API_URL", "http://omni-api:80")
 
-
-class UserMessage(object):
+class UserMessage:
     """Represents an incoming message.
      Includes the channel the responses should be sent to."""
 
@@ -76,6 +82,7 @@ class UserMessage(object):
             self.sender_id = self.DEFAULT_SENDER_ID
 
         self.input_channel = input_channel
+        self.channel_id = metadata.get('channel_id') if metadata else None
 
         self.parse_data = parse_data
         self.metadata = metadata
@@ -97,21 +104,23 @@ def register(
     app.input_channels = input_channels
 
 
-class InputChannel(object):
+class InputChannel:
+    _available_endpoints = None
+
     @classmethod
-    def name(cls):
+    def name(cls) -> Text:
         """Every input channel needs a name to identify it."""
         return cls.__name__
 
     @classmethod
-    def from_credentials(cls, credentials):
+    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> "InputChannel":
         return cls()
 
-    def url_prefix(self):
+    def url_prefix(self) -> Text:
         return self.name()
 
     def blueprint(
-        self, on_new_message: Callable[[UserMessage], Awaitable[None]]
+        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
     ) -> Blueprint:
         """Defines a Sanic blueprint.
         The blueprint will be attached to a running sanic server and handle
@@ -119,7 +128,7 @@ class InputChannel(object):
         raise NotImplementedError("Component listener needs to provide blueprint.")
 
     @classmethod
-    def raise_missing_credentials_exception(cls):
+    def raise_missing_credentials_exception(cls) -> None:
         raise Exception(
             "To use the {} input channel, you need to "
             "pass a credentials file using '--credentials'. "
@@ -137,6 +146,10 @@ class InputChannel(object):
     async def disconnect_session(self):
         pass
 
+    @classmethod
+    def set_available_endpoints(cls, endpoints: AvailableEndpoints):
+        InputChannel._available_endpoints = endpoints
+
     def get_output_channel(self) -> Optional["OutputChannel"]:
         """Create ``OutputChannel`` based on information provided by the input channel.
         Implementing this function is not required. If this function returns a valid
@@ -149,25 +162,71 @@ class InputChannel(object):
         """
         pass
 
+    def get_channel_id(self, channel_name: str):
+        start = time.perf_counter()
+        resp_status = 500
+        channel_id = None
+        delay = 5
+        max_retries = 3
+        retries = 0
+
+        while retries < max_retries:
+            try:
+                with requests.post(
+                    url=f'{OMNI_API_URL}/getOrCreateChannelId',
+                    data={"channelName": channel_name},
+                    timeout=30,
+                ) as res:
+                    resp_status = res.status_code
+                    resp = res.json()
+                    channel_id = resp.get("channelId", None)
+                    if not channel_id and channel_id != 0:
+                        raise Exception(f"Unable to parse channel id from the response. Response: {resp}")
+                    else:
+                        logger.info(f"Channel: {channel_name}, Channel ID: {channel_id}")
+                        break
+            except Exception:
+                retries += 1
+                logger.warning(f"Failed to get channel id. Retrying in {delay}s...", exc_info=True)
+                time.sleep(delay)
+
+        if retries == max_retries:
+            logger.error(f"Failed to get channel id after {max_retries} retries.")
+
+        duration = time.perf_counter() - start
+        OUTGOING_REQUEST_LATENCY_SEC.labels(f'{OMNI_API_URL}/getOrCreateChannelId', resp_status, "POST").observe(
+            duration)
+        return channel_id
+
     async def startTracking(self, usg_msg, csi, channel, sid):
         # make IDP request here and append guid to message
-        host = environ["IDENTITY_PROVIDER_HOST"]
-        port = environ["IDENTITY_PROVIDER_PORT"]
+        idp_endpoint = InputChannel._available_endpoints.idp
 
-        headers = {"Conversation-Id": sid}
+        if idp_endpoint is None:
+            logger.warning(
+                f'[{sid}] - The endpoint of IDP is not configured. Please make sure you add that to your endpoints.yml!')
+            return usg_msg
+
+        idp_url = idp_endpoint.url or "http://localhost:8080"
+        tenant_id = idp_endpoint.kwargs.get('tenant_id', '')
+
+        headers = {"Conversation-Id": sid, "Tenant-Id": tenant_id}
         params = {
             "CSI": csi,
             "channel": channel
         }
         msg = usg_msg
+        start = time.perf_counter()
+        resp_status = 500
         try:
             async with aiohttp.ClientSession() as session:
                 async with session.get(
-                        url='http://' + host + ':' + port + '/startTracking',
-                        params=params,
-                        headers=headers,
-                        timeout=3,
+                    url=f'{idp_url}/startTracking',
+                    params=params,
+                    headers=headers,
+                    timeout=30,
                 ) as res:
+                    resp_status = res.status
                     resp = await res.json()
 
                     # Reformat the initial string
@@ -178,18 +237,36 @@ class InputChannel(object):
                     else:
                         logger.warning(f"[{sid}] - Unable to retrieve guid from IDP: {resp}")
         except Exception as exp:
-            traceback_msg = traceback.format_exc().replace("\n", "")
-            logger.error(f"[{sid}] - Failed to get guid: {traceback_msg}")
+            logger.error(f"[{sid}] - Failed to get guid", exc_info=True)
+
+        duration = time.perf_counter() - start
+        OUTGOING_REQUEST_LATENCY_SEC.labels(f'{idp_url}/startTracking', resp_status, "POST").observe(duration)
+
         return msg
 
+    def get_metadata(self, request: Request) -> Optional[Dict[Text, Any]]:
+        """Extracts additional information from the incoming request.
 
-class OutputChannel(object):
+         Implementing this function is not required. However, it can be used to extract
+         metadata from the request. The return value is passed on to the
+         ``UserMessage`` object and stored in the conversation tracker.
+
+        Args:
+            request: incoming request with the message of the user
+
+        Returns:
+            Metadata which was extracted from the request.
+        """
+        pass
+
+
+class OutputChannel:
     """Output channel base class.
     Provides sane implementation of the send methods
     for text only output channels."""
 
     @classmethod
-    def name(cls):
+    def name(cls) -> Text:
         """Every output channel needs a name to identify it."""
         return cls.__name__
 
@@ -201,7 +278,7 @@ class OutputChannel(object):
                 recipient_id,
                 message.pop("text"),
                 message.pop("quick_replies"),
-                **message
+                **message,
             )
         elif message.get("buttons"):
             await self.send_text_with_buttons(
@@ -256,36 +333,34 @@ class OutputChannel(object):
     ) -> None:
         """Sends an attachment. Default will just post as a string."""
 
-        await self.send_text_message(
-            recipient_id, "Attachment: {}".format(attachment), **kwargs
-        )
+        await self.send_text_message(recipient_id, f"Attachment: {attachment}")
 
     async def send_text_with_buttons(
         self,
         recipient_id: Text,
         text: Text,
         buttons: List[Dict[Text, Any]],
-        **kwargs: Any
+        **kwargs: Any,
     ) -> None:
         """Sends buttons to the output.
         Default implementation will just post the buttons as a string."""
 
-        await self.send_text_message(recipient_id, text, **kwargs)
+        await self.send_text_message(recipient_id, text)
         for idx, button in enumerate(buttons):
             button_msg = cli_utils.button_to_string(button, idx)
-            await self.send_text_message(recipient_id, button_msg, **kwargs)
+            await self.send_text_message(recipient_id, button_msg)
 
     async def send_quick_replies(
         self,
         recipient_id: Text,
         text: Text,
         quick_replies: List[Dict[Text, Any]],
-        **kwargs: Any
+        **kwargs: Any,
     ) -> None:
         """Sends quick replies to the output.
         Default implementation will just send as buttons."""
 
-        await self.send_text_with_buttons(recipient_id, text, quick_replies, **kwargs)
+        await self.send_text_with_buttons(recipient_id, text, quick_replies)
 
     async def send_elements(
         self, recipient_id: Text, elements: Iterable[Dict[Text, Any]], **kwargs: Any
@@ -298,7 +373,7 @@ class OutputChannel(object):
                 title=element.get("title", ""), subtitle=element.get("subtitle", "")
             )
             await self.send_text_with_buttons(
-                recipient_id, element_msg, element.get("buttons", [], **kwargs)
+                recipient_id, element_msg, element.get("buttons", [])
             )
 
     async def send_custom_json(
@@ -307,24 +382,30 @@ class OutputChannel(object):
         """Sends json dict to the output channel.
         Default implementation will just post the json contents as a string."""
 
-        await self.send_text_message(recipient_id, json.dumps(json_message), **kwargs)
+        await self.send_text_message(recipient_id, json.dumps(json_message))
 
 
 class CollectingOutputChannel(OutputChannel):
     """Output channel that collects send messages in a list
     (doesn't send them anywhere, just collects them)."""
 
-    def __init__(self):
+    def __init__(self) -> None:
         self.messages = []
 
     @classmethod
-    def name(cls):
+    def name(cls) -> Text:
         return "collector"
 
     @staticmethod
     def _message(
-        recipient_id, text=None, image=None, video=None, buttons=None, attachment=None, custom=None
-    ):
+        recipient_id: Text,
+        text: Text = None,
+        image: Text = None,
+        video: Text = None,
+        buttons: List[Dict[Text, Any]] = None,
+        attachment: Text = None,
+        custom: Dict[Text, Any] = None,
+    ) -> Dict:
         """Create a message object that will be stored."""
 
         obj = {
@@ -339,13 +420,13 @@ class CollectingOutputChannel(OutputChannel):
         # filter out any values that are `None`
         return utils.remove_none_values(obj)
 
-    def latest_output(self):
+    def latest_output(self) -> Optional[Dict[Text, Any]]:
         if self.messages:
             return self.messages[-1]
         else:
             return None
 
-    async def _persist_message(self, message) -> None:
+    async def _persist_message(self, message: Dict[Text, Any]) -> None:
         self.messages.append(message)  # pytype: disable=bad-return-type
 
     async def send_text_message(
@@ -380,7 +461,7 @@ class CollectingOutputChannel(OutputChannel):
         recipient_id: Text,
         text: Text,
         buttons: List[Dict[Text, Any]],
-        **kwargs: Any
+        **kwargs: Any,
     ) -> None:
         await self._persist_message(
             self._message(recipient_id, text=text, buttons=buttons)
@@ -397,12 +478,12 @@ class QueueOutputChannel(CollectingOutputChannel):
     (doesn't send them anywhere, just collects them)."""
 
     @classmethod
-    def name(cls):
+    def name(cls) -> Text:
         return "queue"
 
     # noinspection PyMissingConstructor
     def __init__(self, message_queue: Optional[Queue] = None) -> None:
-        super(QueueOutputChannel, self).__init__()
+        super().__init__()
         self.messages = Queue() if not message_queue else message_queue
 
     def latest_output(self):
@@ -419,20 +500,23 @@ class RestInput(InputChannel):
     retrieve responses from the agent."""
 
     @classmethod
-    def name(cls):
+    def name(cls) -> Text:
         return "rest"
 
     @staticmethod
     async def on_message_wrapper(
-        on_new_message: Callable[[UserMessage], Awaitable[None]],
+        on_new_message: Callable[[UserMessage], Awaitable[Any]],
         text: Text,
         queue: Queue,
         sender_id: Text,
         input_channel: Text,
+        metadata: Optional[Dict[Text, Any]],
     ) -> None:
         collector = QueueOutputChannel(queue)
 
-        message = UserMessage(text, collector, sender_id, input_channel=input_channel)
+        message = UserMessage(
+            text, collector, sender_id, input_channel=input_channel, metadata=metadata
+        )
         await on_new_message(message)
 
         await queue.put("DONE")  # pytype: disable=bad-return-type
@@ -453,12 +537,13 @@ class RestInput(InputChannel):
         text: Text,
         sender_id: Text,
         input_channel: Text,
+        metadata: Optional[Dict[Text, Any]],
     ) -> Callable[[Any], Awaitable[None]]:
         async def stream(resp: Any) -> None:
             q = Queue()
             task = asyncio.ensure_future(
                 self.on_message_wrapper(
-                    on_new_message, text, q, sender_id, input_channel
+                    on_new_message, text, q, sender_id, input_channel, metadata
                 )
             )
             result = None  # declare variable up front to avoid pytype error
@@ -472,7 +557,9 @@ class RestInput(InputChannel):
 
         return stream  # pytype: disable=bad-return-type
 
-    def blueprint(self, on_new_message: Callable[[UserMessage], Awaitable[None]]):
+    def blueprint(
+        self, on_new_message: Callable[[UserMessage], Awaitable[None]]
+    ) -> Blueprint:
         custom_webhook = Blueprint(
             "custom_webhook_{}".format(type(self).__name__),
             inspect.getmodule(self).__name__,
@@ -480,22 +567,23 @@ class RestInput(InputChannel):
 
         # noinspection PyUnusedLocal
         @custom_webhook.route("/", methods=["GET"])
-        async def health(request: Request):
+        async def health(request: Request) -> HTTPResponse:
             return response.json({"status": "ok"})
 
         @custom_webhook.route("/webhook", methods=["POST"])
-        async def receive(request: Request):
+        async def receive(request: Request) -> HTTPResponse:
             sender_id = await self._extract_sender(request)
             text = self._extract_message(request)
             should_use_stream = rasa.utils.endpoints.bool_arg(
                 request, "stream", default=False
             )
             input_channel = self._extract_input_channel(request)
+            metadata = self.get_metadata(request)
 
             if should_use_stream:
                 return response.stream(
                     self.stream_response(
-                        on_new_message, text, sender_id, input_channel
+                        on_new_message, text, sender_id, input_channel, metadata
                     ),
                     content_type="text/event-stream",
                 )
@@ -505,7 +593,11 @@ class RestInput(InputChannel):
                 try:
                     await on_new_message(
                         UserMessage(
-                            text, collector, sender_id, input_channel=input_channel
+                            text,
+                            collector,
+                            sender_id,
+                            input_channel=input_channel,
+                            metadata=metadata,
                         )
                     )
                 except CancelledError:
