@@ -64,13 +64,13 @@ class TemplatedNaturalLanguageGenerator(NaturalLanguageGenerator):
         else:
             return None
 
-    @timeit
+    # @timeit
     async def generate(
         self,
         template_name: Text,
         tracker: DialogueStateTracker,
         output_channel: Text,
-        **kwargs: Any
+        **kwargs: Any,
     ) -> Optional[Dict[Text, Any]]:
         """Generate a response for the requested template."""
 
@@ -84,7 +84,7 @@ class TemplatedNaturalLanguageGenerator(NaturalLanguageGenerator):
         template_name: Text,
         filled_slots: Dict[Text, Any],
         output_channel: Text,
-        **kwargs: Any
+        **kwargs: Any,
     ) -> Optional[Dict[Text, Any]]:
         """Generate a response for the requested template."""
 
@@ -100,7 +100,7 @@ class TemplatedNaturalLanguageGenerator(NaturalLanguageGenerator):
         self,
         template: Dict[Text, Any],
         filled_slots: Optional[Dict[Text, Any]] = None,
-        **kwargs: Any
+        **kwargs: Any,
     ) -> Dict[Text, Any]:
         """"Combine slot values and key word arguments to fill templates."""
 
@@ -166,7 +166,7 @@ class TemplatedNaturalLanguageGenerator(NaturalLanguageGenerator):
             elif isinstance(v, list):
                 self.iterateNestedList(v, template_vars)
             else:
-                if(k == 'text' or k == 'isVisible'):
+                if k in ['text', 'isVisible', 'file']:
                     v = re.sub(r"{([^\n]+?)}", r"{0[\1]}", v)
                     v = v.format(template_vars)
                     custom[k] = v
