@@ -315,11 +315,20 @@ class MessageProcessor:
             return False
 
     def get_nlu_threshold(self):
-        """Return the NLU confidence threshold"""
-        form_fallback_policy = self.policy_ensemble.look_up_policy(FormFallbackPolicy) if self.policy_ensemble else None
-        if form_fallback_policy:
-            return form_fallback_policy.nlu_threshold or 0.3
+        """
+        Return the NLU confidence threshold. (It only looks for the nlu_threshold from the policy
+        called FormFallbackPolicy. It will not work if users are using something else that is not called
+        FormFallbackPolicy.
+        """
+        if self.policy_ensemble:
+            policies = self.policy_ensemble.policies
+            for p in reversed(policies):
+                if "FormFallbackPolicy" in p.__str__():
+                    return p.nlu_threshold
+
+            return 0.3
         else:
+            logger.warning(f"Unable to get nlu threshold. Default to 0.")
             return 0.0
 
     def get_asr_threshold(self, tracker: DialogueStateTracker):
@@ -410,6 +419,11 @@ class MessageProcessor:
             metadata["type"] = metadata.get("replyType")
         else:
             metadata["type"] = "text"
+        
+        if metadata and metadata.get("message_type"):
+            metadata["message_type"] = metadata.get("message_type")
+        else:
+            metadata["message_type"] = None
 
         if metadata and metadata.get("error_message"):
             metadata["error"] = True
@@ -458,16 +472,22 @@ class MessageProcessor:
         else:
             parse_data = await self._parse_message(message, tracker)
         
+        message.metadata = self.create_metadata(message, tracker)
+        
+        isDTMF = False
+        if message.metadata['message_type']:
+            isDTMF = message.metadata['message_type'] == 'DTMF'
+        
         redacted_text = message.text
         redacted_entities =  copy.deepcopy(parse_data['entities'])
-        if '/greet' not in redacted_text:
+
+        if not isDTMF and '/greet' not in redacted_text:
             redacted_text = UserUttered.redact_msg(redacted_text)
             if redacted_text != message.text:
                 for e in redacted_entities:
                     e['value'] = UserUttered.redact_msg(str(e['value']))
 
         missed_intent = not self._is_nlu_threshold_passed(parse_data, tracker)
-        message.metadata = self.create_metadata(message, tracker)
         
         self.log_user_msg(redacted_text, message.metadata, tracker)
 
