@@ -1,33 +1,37 @@
 import asyncio
+import warnings
 import logging
+import os
 import shutil
 import os
 from functools import partial
 from typing import List, Optional, Text, Union
 
+import json_logging
 from sanic import Sanic
 from sanic_cors import CORS
 from sanic_prometheus import monitor
 
-import rasa.core
+import rasa.core.utils
 import rasa.utils
+import rasa.utils.common
 import rasa.utils.io
-from rasa.core import constants, utils
-from rasa.core.agent import load_agent, Agent
-from rasa.core.channels import BUILTIN_CHANNELS, console
+from rasa import model, server
+from rasa.constants import ENV_SANIC_BACKLOG, ENV_SANIC_WORKERS, DEFAULT_SANIC_WORKERS
+from rasa.core import agent, channels, constants
+from rasa.core.agent import Agent
+from rasa.core.channels import console
 from rasa.core.channels.channel import InputChannel
 from rasa.core.interpreter import NaturalLanguageInterpreter
+from rasa.core.lock_store import LockStore
 from rasa.core.tracker_store import TrackerStore
-from rasa.core.utils import AvailableEndpoints, configure_file_logging
-from rasa.model import get_model_subdirectories, get_model
-from rasa.utils.common import update_sanic_log_level, class_from_module_path
-from rasa.server import add_root_route
+from rasa.core.utils import AvailableEndpoints
 
 logger = logging.getLogger()  # get the root logger
 
 
 def create_http_input_channels(
-    channel: Optional[Text], credentials_file: Optional[Text]
+    channel: Optional[Text], credentials_file: Optional[Text], endpoints: Optional[Text] = None
 ) -> List["InputChannel"]:
     """Instantiate the chosen input channel."""
 
@@ -44,20 +48,22 @@ def create_http_input_channels(
                 "To connect to all given channels, omit the '--connector' "
                 "argument.".format(channel)
             )
-        return [_create_single_channel(channel, all_credentials.get(channel))]
+        return [_create_single_channel(channel, all_credentials.get(channel), endpoints=endpoints)]
     else:
-        return [_create_single_channel(c, k) for c, k in all_credentials.items()]
+        return [_create_single_channel(c, k, endpoints=endpoints) for c, k in all_credentials.items()]
 
 
-def _create_single_channel(channel, credentials):
+def _create_single_channel(channel, credentials, endpoints=None):
     from rasa.core.channels import BUILTIN_CHANNELS
 
     if channel in BUILTIN_CHANNELS:
+        BUILTIN_CHANNELS[channel].set_available_endpoints(endpoints)
         return BUILTIN_CHANNELS[channel].from_credentials(credentials)
     else:
         # try to load channel based on class name
         try:
-            input_channel_class = class_from_module_path(channel)
+            input_channel_class = rasa.utils.common.class_from_module_path(channel)
+            input_channel_class.set_available_endpoints(endpoints)
             return input_channel_class.from_credentials(credentials)
         except (AttributeError, ImportError):
             raise Exception(
@@ -71,14 +77,14 @@ def _create_single_channel(channel, credentials):
 
 def _create_app_without_api(cors: Optional[Union[Text, List[Text]]] = None):
     app = Sanic(__name__, configure_logging=False)
-    add_root_route(app)
-    CORS(app, resources={r"/*": {"origins": cors or ""}}, automatic_options=True)
+    server.add_root_route(app)
+    server.configure_cors(app, cors)
     return app
 
 
 def configure_app(
     input_channels: Optional[List["InputChannel"]] = None,
-    cors: Optional[Union[Text, List[Text]]] = None,
+    cors: Optional[Union[Text, List[Text], None]] = None,
     auth_token: Optional[Text] = None,
     enable_api: bool = True,
     jwt_secret: Optional[Text] = None,
@@ -91,7 +97,7 @@ def configure_app(
     """Run the agent."""
     from rasa import server
 
-    configure_file_logging(logger, log_file)
+    rasa.core.utils.configure_file_logging(logger, log_file)
 
     if enable_api:
         app = server.create_app(
@@ -105,12 +111,12 @@ def configure_app(
         app = _create_app_without_api(cors)
 
     if input_channels:
-        rasa.core.channels.channel.register(input_channels, app, route=route)
+        channels.channel.register(input_channels, app, route=route)
     else:
         input_channels = []
 
     if logger.isEnabledFor(logging.DEBUG):
-        utils.list_routes(app)
+        rasa.core.utils.list_routes(app)
 
     # configure async loop logging
     async def configure_async_logging():
@@ -125,7 +131,7 @@ def configure_app(
             """Small wrapper to shut down the server once cmd io is done."""
             await asyncio.sleep(1)  # allow server to start
             await console.record_messages(
-                server_url=constants.DEFAULT_SERVER_FORMAT.format(port)
+                server_url=constants.DEFAULT_SERVER_FORMAT.format("http", port)
             )
 
             logger.info("Killing Sanic server now.")
@@ -149,11 +155,17 @@ def serve_application(
     endpoints: Optional[AvailableEndpoints] = None,
     remote_storage: Optional[Text] = None,
     log_file: Optional[Text] = None,
+    ssl_certificate: Optional[Text] = None,
+    ssl_keyfile: Optional[Text] = None,
+    ssl_ca_file: Optional[Text] = None,
+    ssl_password: Optional[Text] = None,
 ):
+    from rasa import server
+
     if not channel and not credentials:
         channel = "cmdline"
 
-    input_channels = create_http_input_channels(channel, credentials)
+    input_channels = create_http_input_channels(channel, credentials, endpoints=endpoints)
 
     app = configure_app(
         input_channels,
@@ -169,10 +181,14 @@ def serve_application(
 
     logging.getLogger("engineio").setLevel(logging.WARNING)
     logging.getLogger("socketio").setLevel(logging.ERROR)
+    ssl_context = server.create_ssl_context(
+        ssl_certificate, ssl_keyfile, ssl_ca_file, ssl_password
+    )
+    protocol = "https" if ssl_context else "http"
 
     logger.info(
         "Starting Rasa server on "
-        "{}".format(constants.DEFAULT_SERVER_FORMAT.format(port))
+        "{}".format(constants.DEFAULT_SERVER_FORMAT.format(protocol, port))
     )
 
     app.register_listener(
@@ -180,14 +196,21 @@ def serve_application(
         "before_server_start",
     )
 
-    async def clear_model_files(app: Sanic, _loop: Text) -> None:
+    # noinspection PyUnresolvedReferences
+    async def clear_model_files(_app: Sanic, _loop: Text) -> None:
         if app.agent.model_directory:
-            shutil.rmtree(app.agent.model_directory)
+            shutil.rmtree(_app.agent.model_directory)
 
     async def close_session(app: Sanic, _loop: Text) -> None:
         """Clean up active user sessions before shutting down server"""
         logger.info("Shutting down Sanic server now...")
 
+        if app.session:
+            try:
+                await app.session.close()
+            except Exception:
+                logger.error("Unable to close aiohttp session", exc_info=True)
+
         for input_channel in app.input_channels:
             if input_channel.is_socketio_channel():
                 await input_channel.disconnect_session()
@@ -195,17 +218,29 @@ def serve_application(
         if app.agent.tracker_store.event_broker:
             app.agent.tracker_store.event_broker.flush()
 
-        if app.session:
-            await app.session.close()
-
     app.register_listener(clear_model_files, "after_server_stop")
     app.register_listener(close_session, "before_server_stop")
 
-    update_sanic_log_level(log_file)
-    num_workers = int(os.environ.get("NUM_WORKERS", 1))
+    rasa.utils.common.update_sanic_log_level(log_file)
+    num_workers = int(os.environ.get(ENV_SANIC_WORKERS, DEFAULT_SANIC_WORKERS))
     logger.info("Spawning {} Sanic workers....".format(num_workers))
     monitor(app).expose_endpoint()
-    app.run(host="0.0.0.0", port=port, access_log=False, workers=num_workers)
+    if os.environ.get('ENABLE_JSON_LOGGING', False):
+        json_logging.init_sanic(enable_json=True)
+        json_logging.init_request_instrument(app)
+    app.config['KEEP_ALIVE_TIMEOUT'] = int(os.environ.get("KEEP_ALIVE_TIMEOUT", 5))
+    app.config['KEEP_ALIVE'] = os.environ.get("SANIC_KEEP_ALIVE", True) != "False"
+    app.run(
+        host="0.0.0.0",
+        port=port,
+        access_log=os.environ.get("SANIC_ACCESS_LOG", False),
+        debug=os.environ.get("SANIC_DEBUG_MODE", False),
+        ssl=ssl_context,
+        backlog=int(os.environ.get(ENV_SANIC_BACKLOG, "100")),
+        workers=rasa.core.utils.number_of_sanic_workers(
+            endpoints.lock_store if endpoints else None
+        ),
+    )
 
 
 # noinspection PyUnusedLocal
@@ -220,35 +255,38 @@ async def load_agent_on_start(
 
     Used to be scheduled on server start
     (hence the `app` and `loop` arguments)."""
-    from rasa.core import broker
+    import rasa.core.brokers.utils as broker_utils
 
+    # noinspection PyBroadException
     try:
-        with get_model(model_path) as unpacked_model:
-            _, nlu_model = get_model_subdirectories(unpacked_model)
+        with model.get_model(model_path) as unpacked_model:
+            _, nlu_model = model.get_model_subdirectories(unpacked_model)
             _interpreter = NaturalLanguageInterpreter.create(nlu_model, endpoints.nlu)
     except Exception:
-        logger.debug("Could not load interpreter from '{}'.".format(model_path))
+        logger.debug(f"Could not load interpreter from '{model_path}'.")
         _interpreter = None
 
-    _broker = broker.from_endpoint_config(endpoints.event_broker)
+    _broker = broker_utils.from_endpoint_config(endpoints.event_broker)
     _tracker_store = TrackerStore.find_tracker_store(
         None, endpoints.tracker_store, _broker
     )
+    _lock_store = LockStore.find_lock_store(endpoints.lock_store)
 
     model_server = endpoints.model if endpoints and endpoints.model else None
 
-    app.agent = await load_agent(
+    app.agent = await agent.load_agent(
         model_path,
         model_server=model_server,
         remote_storage=remote_storage,
         interpreter=_interpreter,
         generator=endpoints.nlg,
         tracker_store=_tracker_store,
+        lock_store=_lock_store,
         action_endpoint=endpoints.action,
     )
 
     if not app.agent:
-        logger.warning(
+        warnings.warn(
             "Agent could not be loaded with the provided configuration. "
             "Load default agent without any model."
         )
